# link to my research report: https://hal.inria.fr/hal-03740209v1

x,y = var('x,y')
t1,t2 = var('t1,t2')
tau,v,z = var('tau,v,z')
v1 = var('v1')
z1 = var('z1')
R.<tau,v,z> = ZZ[]
Q.<t1,t2> = RealField(200)[]
Rtauv.<tau,v> = ZZ[]
Rtau.<tau> = ZZ[]
Rv.<v1> = ZZ[]
Qz.<z1> = RR[]


def indice(d,alpha):   #this function returns a list with the indices i1,i2,j such that 0<=i1+i2+d*j <= d*alpha in the order described in the algorithm SLZ
    resu = []
    for k in [0..d*alpha]:
        for j in [0..alpha]:
            i = k-j*d
            if i >= 0:
                for i1 in [0..i]:
                    resu.append((i1,i-i1,j))
    return resu

def nb(d,alpha):   # this function returns the dimension of the lattice
    return 1/12*(2*alpha**3+3*alpha**2+alpha)*d**2+3/4*(alpha**2+alpha)*d+alpha+1

def calcul_erreur_3(f,d):   # this function returns the error for the power function, I worked with d=1,2 and 3, assert d in [1,2,3]
    if d == 1:
        return abs(diff(f,x,2))(1/2,1/2)/2 + abs(diff(f,y,2))(1,1)/2 + abs(diff(diff(f,x,1),y,1))(1,1)
    if d == 2:
        return (abs(diff(f,x,3))/6+abs(diff(f,y,3))/6+abs(diff(diff(f,x,2),y,1))/2+abs(diff(diff(f,x,1),y,2))/2)(1,1)
    if d == 3:
        return (abs(diff(f,x,4))/24+abs(diff(f,y,4))/24+abs(diff(diff(f,x,3),y,1))/6+abs(diff(diff(f,x,2),y,2))/4+abs(diff(diff(f,x,1),y,3))/6)(1/2,1/2)

def bon_i(x0,T,N):   #it returns the integer i such that x_0 is between 1/2+2iT/N et 1/2+2(i+1)T/N
    x0 = x0-1/2
    x0 = x0*N/(2*T)
    return floor(x0)

def SLZ_bivariate_finite_reduce(f,N1,N2,T,M,d,alpha,n,m): # f is the function, N1 is the input precision, N2 is the output precision, n and m are integers such that the study will be made on the square [1/2 + 2nT/N1 , 1/2 + 2(n+1)T/N1[*[1/2 + 2mT/N1 , 1/2 + 2(m+1)T/N1[
    x0 = 1/2+(2*n+1)*T/N1
    y0 = 1/2+(2*m+1)*T/N1
    resu = []
    g(x,y) = N2*f(x/N1+x0,y/N1+y0)
    # Step 1
    ind1 = indice(d,alpha)
    n1 = nb(d,alpha)
    ind2 = indice(1,alpha)
    n2 = nb(1,alpha)
    P = 0
    for i in [0..d]:
        for j in [0..i]:
            di = diff(diff(g,x,j),y,i-j)
            P = P + di(x=0,y=0)*x**j*y**(i-j)/(factorial(j)*factorial(i-j))
    P = Q(P(x=t1,y=t2))
    # Step 2
    eps2 = calcul_erreur_3(f,d)*T**(d+1)/N2**d
    # Step 3
    M1 = floor((1/2)/(2/M+eps2))
    C = (d+1)*(d+2)*M1/2
    P1 = 0
    for i in [0..d]:
        for j in [0..i]:
            P1 = P1 + ZZ(round(C*P[j,i-j]*T**i))*tau**j*v**(i-j)
    e = []
    ge = []
    # Step 4
    for ii in ind1:
        gen = tau**ii[0]*v**ii[1]*z**ii[2]
        e.append(gen)
    # Step 5
    for ii in ind2:
        poly = (T*tau)**ii[0]*(T*v)**ii[1]*(P1+(d+1)*(d+2)/2*z)**ii[2]*C**(alpha-ii[2])
        ge.append(poly)
    #Step 6
    L = matrix(n2,n1)
    for k in [0..(n1-1)]:
        for l in [0..(n2-1)]:
            L[l,k] = ge[l][ind1[k]]
    # Step 7
    V = L.LLL()
    # Step 8
    Q1 = []
    q = []
    for i in [0..(n2-1)]:
        vi = V[i]
        Qi = 0
        qi = vi.norm(1)
        for j in [0..(n1-1)]:
            Qi = Qi + vi[j]*e[j]
        Q1.append(Qi)
        q.append(qi)
    m = []
    for i in [0..(n2-1)]:
        if  q[i]< C**alpha:
            m.append(i)
    # Step 9
    if len(m)<3:
        return 'Fail'
    # Step 10-11-12
    pp = previous_prime(2*T+1)
    tt,vv,vv1 = var('tt,vv,vv1')
    Fptauv.<tt,vv> = GF(pp)[]
    Fptau.<tt> = GF(pp)[]
    Fpv.<vv1> = GF(pp)[]
    p = 0
    i,j,k = m[0],m[1],m[2]
    k1,k2,k3 = 0,1,2
    while p == 0:
        p1 = Fptauv(Q1[i].resultant(Q1[j],z)(tau=tt/T,v=vv/T,z=0))
        p2 = Fptauv(Q1[i].resultant(Q1[k],z)(tau=tt/T,v=vv/T,z=0))
        if p1 != 0 and p2 != 0: # Step 11
            p = Fptau(p1.resultant(p2,vv)(tt=tt,vv=0))
            if p !=0: # Step 12
                racine = p.roots(GF(pp))
                if racine == []:
                    return 'no solution'
                else:
                    for rac in racine:
                        r,mult = rac
                        pp1 = Fpv(p1(tt=r,vv=vv1))
                        pp2 = Fpv(p2(tt=r,vv=vv1))
                        racine1 = pp1.roots(GF(pp))
                        if racine1 == []:
                            return 'no solution'
                        else:
                            for rac1 in racine1:
                                r1,mult1=rac1
                                if pp2(vv1=r1) == 0:
                                    if abs(numerical_approx(g(ZZ(r),ZZ(r1)) - round(g(ZZ(r),ZZ(r1))),200)) < 1/M and abs(ZZ(r))<=T and abs(ZZ(r1))<=T:
                                        if (ZZ(r),ZZ(r1)) not in resu:
                                            resu.append((ZZ(r),ZZ(r1)))
                                    elif abs(numerical_approx(g(ZZ(r)-pp,ZZ(r1)) - round(g(ZZ(r-pp),ZZ(r1))),200)) < 1/M and abs(ZZ(r)-pp)<=T and abs(ZZ(r1))<=T:
                                        if (ZZ(r)-pp,ZZ(r1)) not in resu:
                                            resu.append((ZZ(r)-pp,ZZ(r1)))
                                    elif abs(numerical_approx(g(ZZ(r),ZZ(r1)-pp) - round(g(ZZ(r),ZZ(r1)-pp)),200)) < 1/M and abs(ZZ(r))<=T and abs(ZZ(r1)-pp)<=T:
                                        if (ZZ(r),ZZ(r1)-pp) not in resu:
                                            resu.append((ZZ(r),ZZ(r1)-pp))
                                    elif abs(numerical_approx(g(ZZ(r)-pp,ZZ(r1)-pp) - round(g(ZZ(r)-pp,ZZ(r1)-pp)),200)) < 1/M and abs(ZZ(r)-pp)<=T and abs(ZZ(r1)-pp)<=T:
                                        if (ZZ(r)-pp,ZZ(r1)-pp) not in resu:
                                            resu.append((ZZ(r)-pp,ZZ(r1)-pp))
        if k3 < len(m)-1:
            k3 = k3+1
        elif k2 < len(m)-2:
            k2 = k2+1
            k3 = k2+1
        elif k1 < len(m)-3:
            k1 = k1+1
            k2 = k1+1
            k3 = k2+1
        else:  #it was the last triplet (end of the Step 10)
            return 'Fail'
        i,j,k=m[k1],m[k2],m[k3]
    return resu


# Test double precision
f(x,y) = x**y
x1 = 4783716528592059/9007199254740992
y1 = 17/32

#test1 = SLZ_bivariate_finite_reduce(f,2**53,2**54,2**13,2**49,2,2,bon_i(x1,2**13,2**53),bon_i(y1,2**13,2**53))

# Test single precision
g(x,y) = (x*2**105)**(y*2**(-9))/2
xx1 = 24522110/2**25
yy1 = 31412478/2**25

#test2 = SLZ_bivariate_finite_reduce(g,2**24,2**25,2**5,2**57,2,2,bon_i(xx1,2**5,2**24),bon_i(yy1,2**5,2**24))
